use std::{collections::HashMap, sync::Arc};

use axum::{
    extract::Query,
    http::StatusCode,
    routing::{get, post},
    Extension, Json,
};
use sqlx::SqlitePool;

use crate::responses::registration::RegistrationResponse;
use crate::{
    models::devices::Device,
    responses::{flow::Flows, registration::RegistrationSuccess},
};
use crate::{
    models::users::User,
    requests::registration::RegistrationRequest,
    responses::username_available::UsernameAvailable,
    types::{authentication_data::AuthenticationData, matrix_user_id::UserId},
    Config,
};

use super::errors::{api_error::ApiError, registration_error::RegistrationError};

pub fn routes() -> axum::Router {
    axum::Router::new()
        .route("/r0/login", get(get_login).post(post_login))
        .route("/r0/register", post(post_register))
        .route("/r0/register/available", get(get_username_available))
}

#[tracing::instrument]
async fn get_login() -> Json<Flows> {
    Json(Flows::new())
}

#[tracing::instrument(skip_all)]
async fn post_login(body: String) -> StatusCode {
    dbg!(body);
    StatusCode::INTERNAL_SERVER_ERROR
}

#[tracing::instrument(skip_all)]
async fn get_username_available(
    Extension(config): Extension<Arc<Config>>,
    Extension(db): Extension<SqlitePool>,
    Query(params): Query<HashMap<String, String>>,
) -> Result<Json<UsernameAvailable>, ApiError> {
    let username = params
        .get("username")
        .ok_or(RegistrationError::MissingUserId)?;
    let user_id = UserId::new(username, &config.homeserver_name)
        .ok()
        .ok_or(RegistrationError::InvalidUserId)?;
    let exists = User::exists(&db, &user_id).await?;

    Ok(Json(UsernameAvailable::new(!exists)))
}

#[tracing::instrument(skip_all)]
async fn post_register(
    Extension(config): Extension<Arc<Config>>,
    Extension(db): Extension<SqlitePool>,
    Json(body): Json<RegistrationRequest>,
) -> Result<Json<RegistrationResponse>, ApiError> {
    body.auth()
        .ok_or(RegistrationError::AdditionalAuthenticationInformation)?;

    let (user, device) = match &body.auth().expect("must be Some") {
        AuthenticationData::Password(auth_data) => {
            let username = body.username().ok_or(RegistrationError::MissingUserId)?;
            let user_id = UserId::new(username, &config.homeserver_name)
                .ok()
                .ok_or(RegistrationError::InvalidUserId)?;

            User::exists(&db, &user_id)
                .await?
                .then(|| ())
                .ok_or(RegistrationError::UserIdTaken)?;

            let password = auth_data.password();

            let display_name = match body.initial_device_display_name() {
                Some(display_name) => display_name.as_ref(),
                None => "Random displayname",
            };

            let user = User::create(&db, &user_id, &user_id.to_string(), password).await?;
            let device = Device::create(&db, &user, "test", display_name).await?;

            (user, device)
        }
    };

    if body.inhibit_login().unwrap_or(false) {
        let resp = RegistrationSuccess::new(None, device.device_id(), user.user_id());

        Ok(Json(RegistrationResponse::Success(resp)))
    } else {
        let session = device.create_session(&db).await?;
        let resp =
            RegistrationSuccess::new(Some(session.value()), device.device_id(), user.user_id());

        Ok(Json(RegistrationResponse::Success(resp)))
    }
}
